function Buckley_Leverett_MatMOL_DynamicRegridding(eps0,n0,Reltol,Abstol)

%**************************************************************************
% The MatMol Group (2009)
% Function....... : Transient simulation of the classic 1D Buckley-Leverett
%                   equation for oil reservoirs
% Technique...... : Solution with the MatMOL toolbox and dynamic regridding
%                   MOVGRID
% Input...........: eps0:     capilary pressure term
%                   n0:       number of vertex centered discritisation points
%                   RelTol:   Relatitve integration tolerance
%                   AbsTol:   Absolute integration tolerance
%**************************************************************************
close all

global zL zR n ne X eq choice bloc0 bloc1 bloc2 bloc3
global alpha kappa mu tau
global u z dltz mon dltu x
global eps
     
if nargin < 1, eps0     = 0.001;    end  
if nargin < 2, n0       = 101;      end  
if nargin < 3, Reltol   = 1.0e-6;   end  
if nargin < 4, Abstol   = 1.0e-6;   end     

tic % initialise clock


% 1 Parameters
% ************

% Process
eps     =   eps0;               % [-]       capillary pressure term
clear eps0

% Spatial grid
n   	=   n0-2;               % [-]       # of moving grid points
zL      =   0.0;                % [-]       left border
zR      =   1.0;                % [-]       right border
dz      =   (zR-zL)/(n+1);      % [-]       spatial discretisation length
z       =   [zL:dz:zR]';        % [-]       position vector

% Note:
% If n0 is the original number of discretisation points (i.e., dim(z)), 
% n is the number of moving grid points. (z0 and zL are fixed.)

% Temporal grid
t_start =   0.0;                % [-]       initial time
t_end   =   0.4;                % [-]       final time
dt_plot =   0.01;               % [-]       time interval for plots
t_plot  =   t_start:dt_plot:t_end;  % [-]   time points to be plotted

% Parameters of the moving grid (MOVGRID)
ne      =   1;                  % [-]       nr of PDEs
alpha   =   1.0;                % [-]
kappa   =   2.0;                % [-]
mu      =   kappa*(kappa+1);    % [-]
tau     =   0.05;               % [-]
choice  =   'derfl1-1';         % [-]       monitoring function choice


% 2 Numerical computation
% ***********************  

% 2a Initial conditions
S0 = max(zeros(n+2,1),interp1([0 1/3 1],[1 0 0],z));

% 2b Initialisation global variables for JPat
X = zeros(n*(ne+1),n*(ne+1));
eq(:,:,1) = [-2 2 1];

% 2c Initialisation general blocs of MvPattern
bloc0 = zeros(ne+1,ne+1);
bloc1 = bloc0;
bloc1(ne+1,:) = 1;
bloc2 = eye(ne+1);
bloc2(:,ne+1) = 1;
bloc2(ne+1,:) = 1;
bloc3 = bloc0;
bloc3(ne+1,ne+1) = 1;

% 2d Adapt the initial grid
% (This step can be omitted, if no regridding of the initial profiles is 
% requested.)

% Compute the spatial smoothing matrix
Ainit = diag((1+2*mu)*ones(n+1,1),0)+diag(-mu*ones(n,1),1)+diag(-mu*ones(n,1),-1);
Ainit(1,1)     = Ainit(1,1)-mu;
Ainit(n+1,n+1) = Ainit(n+1,n+1)-mu;

% Iterative initial grid adaptation
for iad=1:6
    % Compute the monitoring function mon(i)
    mon = monitor(S0,z,choice,t_plot);

    % Compute the new grid
    suminv=0;
    for i=1:n+1,
        suminv=suminv+1/mon(i);
    end
    for i=1:n+1,
        g(i)=(zR-zL)/(mon(i)*suminv);
    end
    delz=Ainit\g';
    for i=2:n+1,
        z(i)=z(i-1)+delz(i-1);
    end
    % New initial dependent variables S0
    S0 = max(zeros(n+2,1),interp1([0 1/3 1],[1 0 0],z));
end

% Global initial condition
for j=1:ne,
    x(j:ne+1:n*(ne+1))=S0(2:n+1,j);
end
x(ne+1:ne+1:n*(ne+1))=z(2:n+1);

% 2e Solution of the DAEs
options = odeset('RelTol',Reltol,'AbsTol',Abstol,'Mass',@mass,...
         'MStateDependence','strong','JPattern',JPat,...
         'MvPattern',MvPat,'MassSingular','no');
sol  = ode15s(@ODE_BL_MatMOL_DynamicRegridding,t_plot,x,options);
t    = t_plot;                  % time points to be plotted
sol2 = deval(sol,t);            % evaluate solution at time points
y    = sol2';
clear sol2

% 2f Properties
% Separate the dependent variables and the node positions
Z = zeros(n+2,1);
S = zeros(n+2,ne);
for i=1:length(t),
    for j=1:n,
        Z(j+1,i) = y(i,(ne+1)*j);
        S(j+1,i) = y(i,(ne+1)*j-1);
    end
    Z(1,i)   = zL;
    Z(n+2,i) = zR;
    S(1,i)   = 1.0;
    S(n+2,i) = 0.0;    
end

% 2g Integrator statisctics
Stats   = sol.stats; % solution statistics
CPUtime = toc;       % read clock
clear sol

disp('Integration statistics:')
fprintf('  %g successful steps\n', Stats.nsteps);
fprintf('  %g failed attempts\n',  Stats.nfailed);
fprintf('  %g function evaluations\n', Stats.nfevals);
fprintf('  %g partial derivatives\n',  Stats.npds);
fprintf('  %g LU decompositions\n', Stats.ndecomps);
fprintf('  %g solutions of linear systems\n', Stats.nsolves);
fprintf('  %g s (total) computation time\n',  CPUtime);


% 3 Plot results
% **************

figure(1)
subplot('position',[0.1 0.3 0.8 0.6])
plot(Z,S,'k',Z,S,'.k')
ylabel('Saturation [-]','Fontsize',18)
title(['MatMOL Dynamic Regridding: \epsilon=' num2str(eps)],'Fontsize',18)
axis([zL zR -0.1 1.1])
set(gca,'Fontsize',16)
    
subplot('position',[0.1 0.1 0.8 0.15])
for i=1:length(t),
    plot(Z(:,i),t_plot(i)*ones(1,n+2),'.k')
    hold on
end
xlabel('Space [-]','fontsize',18);
ylabel('Time [-]','Fontsize',18)
axis([zL zR t_start t_end])
set(gca,'Fontsize',16)

end % End of function Buckley_Leverett_MatMOL_DynamicRegridding
% *************************************************************************

function xt = ODE_BL_MatMOL_DynamicRegridding(t,x)
% Compute the right vector of the global ODEs system

% The following code implements the right vector b(x,t) of the system
%
%            A(x,t)xdot = b(x,t)
%
% with
%          1   2         ne      1   2         ne            1         ne
%    x = [u   u   ...   u   z   u   u   ...   u   z   ...   u   ...   u   z ]'
%          1   1         1   1   2   2         2   2         n         n   n
%
% and where the meshpoints 1, 2, ... n are the interiors nodes. Nevertheless, b(x,t) uses also
%
% the boundary points and values z , z   , u  , u   .
%                                 0   n+1   0    n+1
%
% The code implements first  a) the monitor function M(i) :
%
%                                 j     j 2
%                               (u   - u )
%                         1  ne   i+1   i
%    M(i) = sqrt{alpha + -- sum --------   } , i = 0, ..., n
%                        ne j=1           2
%                               (z   - z )
%                                 i+1   i
%
%                            b) the "deltaz" vector dltz(i) :
%
%    dltz(i) = z   - z  , i = -1, ..., n+1 
%               i+1   i
%
% with   dltz(-1) = dltz(0)  and   dltz(n+1) = dltz(n).
%
% Following elements of b(x,t) are then implemented :
%
%    a) elements from the moving grid equation :
%
%            1         mu        1+2mu       mu          1         mu        1+2mu       mu
%    g(i) = --- [ - --------- + ------- - -------- ] - ----- [ - ------- + --------- - -------- ]
%           M(i)    dltz(i+1)   dltz(i)   dltz(i-1)    M(i-1)    dltz(i)   dltz(i-1)   dltz(i-2)
%
%    b) elements from the PDEs :
%
%    udot(i,j)   i = 1, ..., n       j = 1, ..., ne.
%
% Structure of b(x,t) : b(x,t) is (ne+1)*n vector :
%
%                udot(1,1)
%
%                udot(1,2)
%
%                . . .
%
%                udot(1,ne)
%
%                g(1)
%
%                udot(2,1)
%
%                udot(2,2)
%
%                . . .
%    b(x,t) =
%                udot(2,ne)
%
%                g(2)
%
%                . . . . . . . .
%
%                udot(n,1)
%
%                udot(n,2)
%
%                . . .
%
%                udot(n,ne)
%
%                g(n)

global zL zR n ne D1
global alpha kappa mu tau
global u z dltz mon dltu
global mo_inv dz_inv
global eps
global choice

% Separate dependent variables and node positions and implement the BCs
[u z] = BC_DynGrid(t,x);

% Compute the monitoring function mon(i)
mon = monitor(u,z,choice,t);

% Compute the right vector g(n) of the moving grid equation
g = rightmon(mon,dltz,mu);
    
% Diffusion term
D1=three_point_centered_D1(z);
udz=D1*u;
fd=4*u.*(1-u).*udz;
fdz=D1*fd;

% Convection term
uz=kurg_centred_slope_limiter_fz(ne,n+2,z',t,u,'Flux_DynGrid',...
                            'dFlux_DynGrid_dx');

% Temporal derivatives
udot = -uz + eps*fdz;

udot(1)    = 0;
udot(n+2)  = 0;

% Assemble the global right vector
for j=1:ne,
    xt(j:ne+1:n*(ne+1),1)=udot(2:n+1,j);
end
xt(ne+1:ne+1:n*(ne+1),1)=g;

end % End of function ODE_BL_MatMOL_DynamicRegridding
% *************************************************************************
%
% function Flux_DynGrid(n,ne,t,X)
% See separate m-file in the same directory
%
%**************************************************************************
%
% function dfxdx = dFlux_DynGrid_dx(n,ne,X)
% See separate m-file in the same directory
%
%**************************************************************************
%
% function BC_DynGrid(t,X)
% See separate m-file in the same directory
%
%**************************************************************************